@@ -19,6 +19,11 @@ import hashlib
 import base64
 from collections import deque
 import threading
+import hmac
+from typing import Any
+from typing import Optional
+import mimetypes
+from urllib.parse import urlencode
 
 # Global lock untuk sinkronisasi akses ke cache
 cache_lock = threading.Lock()
@@ -35,7 +40,34 @@ duplicate_time_window = 10  # dalam detik
 
 # Secret Key untuk API Key
 SECRET_KEY = "MySatnusa"
-API_KEY_EXPIRATION = 86400 
+
+# Konfigurasi batas ukuran
+MAX_JSON_SIZE_MB = 4
+MAX_FILE_SIZE_MB = 2
+MAX_TOTAL_SIZE_MB = 10
+MAX_FILES_ALLOWED = 20
+ALLOWED_FILE_EXTENSIONS = {"jpg", "jpeg", "png", "pdf", "txt"}
+ALLOWED_MIME_TYPES = {
+    "jpg": "image/jpeg",
+    "png": "image/png",
+    "pdf": "application/pdf",
+    "txt": "text/plain"
+}
+# Daftar User-Agent yang dikenal sebagai bot sah (opsional)
+ALLOWED_BOTS = {
+    "Googlebot": "Googlebot",
+    "Bingbot": "bingbot",
+    "YandexBot": "YandexBot",
+    # Tambahkan bot lain yang diizinkan sesuai kebutuhan
+}
+
+# Daftar pola User-Agent mencurigakan
+SUSPICIOUS_UA_PATTERNS = [
+    r"(?i)\b(bot|crawl|spider|slurp|http|scrap|curl|wget|python-requests|java|node|phantomjs)\b",  # Bot/scraper umum
+    r"(?i)\b(headless|automation|test|fake|unknown)\b",  # Indikator headless atau tes
+    r"^[0-9]+$",  # User-Agent hanya angka (sangat tidak wajar)
+    r"^\s*$",  # Kosong atau hanya whitespace
+]
  
 # Helper: Execute query 
 def execute_query(sql_query, params=None, db_alias='default'):
@@ -835,10 +867,10 @@ def validate_request(data, rules):
 def validate_method(
     request,
     required_method,
+    require_api_key=False,
     rate_limit=60,
     time_window=60,
     method_specific=True,
-    require_api_key=False,
     block_bots=True
 ):
     """
@@ -862,6 +894,33 @@ def validate_method(
     # 1. Validasi Method
     if method != required_method.upper():
         raise ValueError(f"Method {method} not allowed. Use {required_method}.")
+    
+    # VALIDATE REQUEST [START]
+    # VALIDATE REQUEST [START]
+    # VALIDATE REQUEST [START]
+
+    # 2. Cek apakah Content-Type valid (hanya menerima JSON atau multipart/form-data)
+    content_type = request.headers.get("Content-Type", "")
+    if not any(ct in content_type for ct in ["application/json", "multipart/form-data"]):
+        raise ValueError("Invalid Content-Type. Only application/json or multipart/form-data is allowed.")
+
+    # 3. Batasi ukuran request body
+    max_size_bytes = MAX_TOTAL_SIZE_MB * 1024 * 1024  # 10 MB untuk multipart
+    if request.META.get('CONTENT_LENGTH') and int(request.META.get('CONTENT_LENGTH')) > max_size_bytes:
+        raise ValueError("Payload too large. Maximum allowed size is 10MB.")
+
+    # 4. Validasi header User-Agent
+    validate_user_agent(request)
+
+    # 5. Validasi request body berdasarkan tipe content
+    if "application/json" in content_type:
+        validate_json_payload(request)
+    elif "multipart/form-data" in content_type:
+        validate_file_upload(request)
+        
+    # VALIDATE REQUEST [END]
+    # VALIDATE REQUEST [END]
+    # VALIDATE REQUEST [END]
 
     # 2. Rate Limiting per Method dan IP berdasarkan endpoint
     with cache_lock:
@@ -890,7 +949,7 @@ def validate_method(
             ip_cache.append(now)
 
     # 3. Duplicate Request Limiting
-    request_signature = generate_request_signature(request, endpoint)
+    # request_signature = generate_request_signature(request, endpoint)
     with cache_lock:
         endpoint_signatures = REQUEST_SIGNATURES.setdefault(endpoint, {})
         ip_signatures = endpoint_signatures.setdefault(client_ip, deque(maxlen=duplicate_limit))
@@ -903,27 +962,22 @@ def validate_method(
     # 4. Validasi API Key (Opsional)
     if require_api_key:
         api_key = request.headers.get("X-API-KEY")
-        timestamp = request.headers.get("X-TIMESTAMP")
 
-        if not api_key or not timestamp:
-            raise ValueError("Unauthorized: Missing API Key or Timestamp")
+        # Pastikan API key ada
+        if not api_key:
+            raise ValueError("401 Unauthorized: Missing API Key")
+        
+        # Menentukan periode 6 bulan
+        current_month = datetime.utcnow().month
+        year = datetime.utcnow().year
+        period = f"{year}-{'01' if current_month <= 6 else '07'}"  # Januari-Juni -> periode 01, Juli-Desember -> periode 07
 
-        try:
-            request_time = int(timestamp)
-            current_time = int(time.time())
+        # Membuat API key yang diharapkan
+        expected_api_key = hmac.new(SECRET_KEY.encode(), period.encode(), hashlib.sha256).hexdigest()
 
-            if abs(current_time - request_time) > API_KEY_EXPIRATION:
-                raise ValueError("Unauthorized: Expired or invalid timestamp")
-
-            # Generate API Key untuk bulan ini
-            month_string = datetime.datetime.utcfromtimestamp(request_time).strftime("%Y-%m")
-            expected_api_key = generate_monthly_api_key(month_string)
-
-            if not hmac.compare_digest(api_key, expected_api_key):
-                raise ValueError("Unauthorized: Invalid API Key")
-
-        except ValueError:
-            raise ValueError("Unauthorized: Invalid timestamp format")
+        # Bandingkan API key yang diterima dengan yang diharapkan
+        if api_key != expected_api_key:
+            raise ValueError("Unauthorized: Invalid API Key")
 
     # 5. Blokir User-Agent yang Mencurigakan (Anti Bot & Scraper)
     if block_bots:
@@ -956,10 +1010,165 @@ def generate_request_signature(request, endpoint):
     hash_digest = hashlib.sha256(data.encode()).digest()
     return base64.b64encode(hash_digest).decode()
 
-def generate_monthly_api_key(month_string):
-    """Membuat API key berdasarkan bulan"""
-    return hmac.new(SECRET_KEY.encode(), month_string.encode(), hashlib.sha256).hexdigest()
- 
+# Fungsi untuk memvalidasi payload JSON
+def validate_json_payload(request):
+    """Validasi payload JSON dari request."""
+
+    if request.body:
+        try:
+            data = json.loads(request.body)
+        except json.JSONDecodeError:
+            raise ValueError("Invalid JSON format.")
+        except Exception:
+            raise ValueError("Invalid request payload.")
+
+        # Batasi ukuran JSON payload
+        if len(request.body) > (MAX_JSON_SIZE_MB * 1024 * 1024):
+            raise ValueError(f"JSON payload too large. Maximum allowed size is {MAX_JSON_SIZE_MB}MB.")
+
+        # Deteksi input mencurigakan secara rekursif
+        if contains_malicious_input(data):
+            raise ValueError("Potential hacking detected.")
+
+def validate_file_upload(request):
+    """Validasi file upload dan field formulir dari request multipart/form-data."""
+
+    # 1. Validasi field formulir (request.POST)
+    form_data = request.POST
+    if contains_malicious_input(form_data):
+        raise ValueError("Potential hacking detected in form fields.")
+
+    # 2. Validasi file yang diunggah
+    files = []
+    for key in request.FILES:
+        files.extend(request.FILES.getlist(key))
+
+    if len(files) > MAX_FILES_ALLOWED:
+        raise ValueError(f"Too many files uploaded. Maximum {MAX_FILES_ALLOWED} files are allowed.")
+
+    total_size = 0
+    for uploaded_file in files:
+        file_size_mb = uploaded_file.size / (1024 * 1024)
+        total_size += file_size_mb
+
+        # Validasi ukuran per file
+        if file_size_mb > MAX_FILE_SIZE_MB:
+            raise ValueError(f"File '{uploaded_file.name}' is too large. Maximum allowed size is {MAX_FILE_SIZE_MB}MB.")
+
+        # Validasi ekstensi file
+        file_extension = uploaded_file.name.split(".")[-1].lower()
+        if file_extension not in ALLOWED_FILE_EXTENSIONS:
+            raise ValueError(f"File '{uploaded_file.name}' has an invalid extension. Allowed: {', '.join(ALLOWED_FILE_EXTENSIONS)}.")
+
+        # Validasi MIME type menggunakan mimetypes
+        mime_type, _ = mimetypes.guess_type(uploaded_file.name, strict=True)
+        expected_mime = ALLOWED_MIME_TYPES.get(file_extension)
+        if mime_type != expected_mime:
+            raise ValueError(f"File '{uploaded_file.name}' has an invalid MIME type: {mime_type or 'unknown'}. Expected: {expected_mime}.")
+
+        # Validasi isi file untuk file teks (opsional)
+        if file_extension == "txt":
+            try:
+                file_content = uploaded_file.read().decode("utf-8")
+                if contains_malicious_input(file_content):
+                    raise ValueError(f"File '{uploaded_file.name}' contains potentially malicious content.")
+            except UnicodeDecodeError:
+                raise ValueError(f"File '{uploaded_file.name}' is not a valid text file.")
+            finally:
+                uploaded_file.seek(0)  # Reset pointer
+
+    # Validasi total ukuran
+    if total_size > MAX_TOTAL_SIZE_MB:
+        raise ValueError(f"Total file upload size exceeds {MAX_TOTAL_SIZE_MB}MB.")
+
+# Daftar pola berbahaya (dari kode Anda sebelumnya)
+suspicious_patterns = [
+    # XSS Injection
+    r"<script\b[^>]*>.*?</script>",  # Hanya deteksi <script> dengan konten berbahaya
+    r"(?i)\b(on\w+=['\"][^'\"]*['\"])",  # Event handler seperti onload, onclick
+    r"(?i)<iframe\b[^>]*src=['\"][^'\"]*['\"][^>]*>.*?</iframe>",  # Iframe dengan src mencurigakan
+    r"(?i)<meta\b[^>]*http-equiv=['\"]refresh['\"][^>]*content=['\"][^'\"]*['\"][^>]*>",  # Meta refresh redirect
+    r"(?i)<img\b[^>]*src=['\"]javascript:[^'\"]*['\"][^>]*>",  # Img dengan javascript: URI
+    
+    # SQL Injection
+    r"(?i)\b(union\s+select|drop\s+table|--\s+|/\*|\*/|;\s*xp_cmdshell|exec\s+xp_)\b|#(?=\s*(select|insert|update|delete|drop))",  # Perintah SQL berbahaya
+    r"(?i)\b(select\s+.*?from|insert\s+into|update\s+.*?set|delete\s+from)\s*['\";]",  # Query SQL dengan tanda berbahaya
+    r"(?i)\b(load_file|outfile|dumpfile|sleep\(\d+\)|benchmark\(\d+,\d+\))\b",  # Fungsi SQL berbahaya
+    
+    # JavaScript Injection
+    r"(?i)\b(alert|confirm|prompt|document\.cookie|eval|setTimeout|setInterval|Function)\s*\(",  # Fungsi JS berbahaya dengan pemanggilan
+    r"(?i)\b(fetch|XMLHttpRequest|webkitRequestFileSystem)\s*\(",  # API berbahaya dengan pemanggilan
+    r"(?i)\b(window\.location|history\.go|localStorage|sessionStorage)\s*[=\.]",  # Manipulasi browser
+    
+    # Dangerous Encodings & URI Schemes
+    r"(?i)%3Cscript%3E|%3Ciframe%3E|%3Cimg%20src=['\"]javascript:|%3Cbody%20onload|%3Conerror=",  # Encoded attacks
+    r"(?i)\b(data:text/html|javascript:|vbscript:|mocha:|livescript:|file://)\b",  # URI berbahaya
+    
+    # NoSQL Injection & Command Injection
+    r"(?i)\b(\$ne|\$eq|\$gte|\$lte|\$regex|\$where)\s*[:=]",  # MongoDB operator mencurigakan
+    r"(?i)\b(cat\s+/etc/passwd|ls\s+-la|whoami|id|uname\s+-a|rm\s+-rf|chmod\s+777)\b",  # Perintah sistem
+    
+    # Path Traversal Attack
+    r"(?i)\b(\.\./|\.\.\\|/etc/passwd|c:\\windows\\system32)\b",  # Path traversal
+]
+
+# Fungsi rekursif untuk memeriksa input berbahaya
+def contains_malicious_input(data: Any) -> bool:
+    """
+    Cek apakah payload mengandung karakter berbahaya secara rekursif.
+    Args:
+        data: Data yang akan diperiksa (bisa dict, list, atau str).
+    Returns:
+        bool: True jika ada input berbahaya, False jika aman.
+    """
+    if isinstance(data, dict):
+        return any(contains_malicious_input(value) for value in data.values())
+    elif isinstance(data, list):
+        return any(contains_malicious_input(item) for item in data)
+    elif isinstance(data, str):
+        sanitized_value = data.strip()
+        for pattern in suspicious_patterns:
+            if re.search(pattern, sanitized_value):
+                return True
+    return False
+
+def validate_user_agent(request) -> None:
+    """
+    Validasi header User-Agent untuk mendeteksi permintaan mencurigakan.
+    Args:
+        request: Request object Django.
+    Raises:
+        ValueError: Jika User-Agent mencurigakan atau tidak valid.
+    """
+    user_agent: str = request.headers.get("User-Agent", "").strip()
+    
+    # 1. Cek apakah User-Agent ada dan memiliki panjang minimal
+    if not user_agent:
+        raise ValueError("Missing User-Agent. Request seems suspicious.")
+    if len(user_agent) < 10:
+        # Izinkan pengecualian untuk bot yang sah jika diperlukan
+        if not any(bot.lower() in user_agent.lower() for bot in ALLOWED_BOTS.values()):
+            raise ValueError(f"User-Agent too short ({len(user_agent)} chars). Request seems suspicious.")
+
+    # 2. Cek pola User-Agent mencurigakan
+    for pattern in SUSPICIOUS_UA_PATTERNS:
+        if re.search(pattern, user_agent):
+            # Izinkan bot yang dikenal (opsional)
+            if any(bot.lower() in user_agent.lower() for bot in ALLOWED_BOTS.values()):
+                return  # Bot sah, lanjutkan
+            raise ValueError(f"Suspicious User-Agent detected: {user_agent}")
+
+    # 3. Cek apakah User-Agent terlalu generik atau tidak wajar
+    if user_agent.lower() in ["mozilla", "browser", "default", "user-agent"]:
+        raise ValueError("Generic User-Agent detected. Request seems suspicious.")
+
+    # 4. (Opsional) Cek header tambahan untuk konsistensi
+    accept_header = request.headers.get("Accept", "")
+    if not accept_header and "html" not in accept_header.lower():
+        # Jika tidak ada Accept header atau tidak menerima HTML, kemungkinan bukan browser
+        if not any(bot.lower() in user_agent.lower() for bot in ALLOWED_BOTS.values()):
+            raise ValueError("Inconsistent headers with User-Agent. Request seems suspicious.")
+
 def log_exception(request, exception):
     try:
         # Parse traceback to get the file and line where the error occurred
@@ -1226,4 +1435,263 @@ def generate_excel_from_template(data, url=None, output_file_name="dashboard_rep
         return output_file
  
     except Exception as e:
-        raise Exception(f"Error saving Excel file: {e}")
\ No newline at end of file
+        raise Exception(f"Error saving Excel file: {e}")
+    
+def execute_query_with_pagination(
+    request,
+    sql_query,
+    params = None,
+    db_alias = 'default',
+    default_page = 1,
+    default_per_page = 10,
+    order_by = None
+) -> dict:
+    """
+    Executes a raw SQL query with pagination on the specified database and returns paginated results.
+ 
+    Args:
+        request: HttpRequest object to get pagination parameters and build URLs.
+        sql_query: The base SQL query to execute.
+        params: List of parameters to safely pass to the query.
+        db_alias: The alias of the database to use.
+        default_page: Default page number if not specified in request.
+        default_per_page: Default items per page if not specified in request.
+        order_by: Optional ORDER BY clause for sorting results.
+ 
+    Returns:
+        dict: A dictionary containing paginated data and metadata:
+            - rows: List of result rows
+            - current_page: Current page number
+            - page_size: Items per page
+            - total_pages: Total number of pages
+            - total_rows: Total number of rows
+            - last_page: Last page number
+            - next_page_url: URL for next page (or None)
+            - prev_page_url: URL for previous page (or None)
+    """
+    try:
+        # Initialize variables
+        params = params or []
+        pagination = {}
+        
+        # Get pagination parameters from request
+        try:
+            page = int(request.GET.get("page", default_page))
+            per_page = int(request.GET.get("page_size", default_per_page))
+        except ValueError:
+            page = default_page
+            per_page = default_per_page
+        
+        offset = (page - 1) * per_page
+        
+        # Count total rows using execute_query
+        count_query = f"SELECT COUNT(*) as total FROM ({sql_query}) AS total_data"
+        total_rows = execute_query(count_query, params, db_alias)[0]["total"]
+        
+        # Calculate pagination metadata
+        total_pages = (total_rows + per_page - 1) // per_page
+        last_page = total_pages
+        
+        # Prepare paginated query
+        paginated_params = params.copy()
+        paginated_query = sql_query
+        if order_by:
+            paginated_query += f" {order_by}"
+        paginated_query += " LIMIT %s OFFSET %s"
+        paginated_params.extend([per_page, offset])
+        
+        # Execute paginated query using execute_query
+        data = execute_query(paginated_query, paginated_params, db_alias)
+        
+        # Build pagination URLs
+        base_url = request.build_absolute_uri(request.path)
+        query_params = request.GET.copy()
+        
+        next_page_url = None
+        prev_page_url = None
+        
+        if page < total_pages:
+            query_params['current_page'] = str(page + 1)
+            next_page_url = f"{base_url}?{urlencode(query_params)}"
+        
+        if page > 1:
+            query_params['current_page'] = str(page - 1)
+            prev_page_url = f"{base_url}?{urlencode(query_params)}"
+        
+        # Construct response
+        pagination = {
+            "current_page": page,
+            "page_size": per_page,
+            "total_pages": total_pages,
+            "total_rows": total_rows,
+            "last_page": last_page,
+            "next_page_url": next_page_url,
+            "prev_page_url": prev_page_url,
+            "rows": data,
+        }
+        
+        return pagination
+    
+    except Exception as e:
+        raise Exception(f"Error executing paginated query on database '{db_alias}': {e}")
+    
+def get_data_with_pagination(
+    request,
+    table_name,
+    filters = None,
+    search = None,
+    search_columns = None,
+    columns = '*',
+    order_by = None,
+    db_alias = 'default',
+    default_page = 1,
+    default_per_page = 10
+) -> dict:
+    """
+    Helper to read data from a table with pagination, supporting various operators like !=, IN, NOT IN, and NULL handling.
+ 
+    Args:
+        request: HttpRequest object to get pagination parameters and build URLs.
+        table_name: The name of the table.
+        filters: Conditions to filter data.
+        search: Search keyword.
+        search_columns: Columns to perform search on.
+        columns: Columns to select, default is '*'.
+        order_by: Column(s) to order by.
+        db_alias: The database alias to use (default is 'default').
+        default_page: Default page number if not specified in request.
+        default_per_page: Default items per page if not specified in request.
+ 
+    Returns:
+        dict: A dictionary containing paginated data and metadata:
+            - rows: List of result rows
+            - current_page: Current page number
+            - page_size: Items per page
+            - total_pages: Total number of pages
+            - total_rows: Total number of rows
+            - last_page: Last page number
+            - next_page_url: URL for next page (or None)
+            - prev_page_url: URL for previous page (or None)
+    """
+    try:
+        # Initialize variables
+        pagination = {}
+        values = []
+        conditions = []
+        
+        # Get pagination parameters from request
+        try:
+            page = int(request.GET.get("page", default_page))
+            per_page = int(request.GET.get("page_size", default_per_page))
+        except ValueError:
+            page = default_page
+            per_page = default_per_page
+        
+        offset = (page - 1) * per_page
+        
+        # Build base query
+        if isinstance(columns, list):
+            columns = ', '.join(columns)
+        sql_query = f"SELECT {columns} FROM {table_name}"
+        
+        # Add filters
+        if filters:
+            for key, value in filters.items():
+                if "__ne" in key:
+                    column = key.replace("__ne", "")
+                    if value is None:
+                        conditions.append(f"{column} IS NOT NULL")
+                    else:
+                        conditions.append(f"{column} <> %s")
+                        values.append(value)
+                elif "__in" in key:
+                    column = key.replace("__in", "")
+                    if isinstance(value, (list, tuple)):
+                        conditions.append(f"{column} IN ({', '.join(['%s'] * len(value))})")
+                        values.extend(value)
+                    else:
+                        conditions.append(f"{column} = %s")
+                        values.append(value)
+                elif "__not_in" in key:
+                    column = key.replace("__not_in", "")
+                    if isinstance(value, (list, tuple)):
+                        conditions.append(f"{column} NOT IN ({', '.join(['%s'] * len(value))})")
+                        values.extend(value)
+                    else:
+                        conditions.append(f"{column} <> %s")
+                        values.append(value)
+                else:
+                    if value is None:
+                        conditions.append(f"{key} IS NULL")
+                    else:
+                        conditions.append(f"{key} = %s")
+                        values.append(value)
+        
+        # Add search conditions
+        if search and search_columns:
+            search_conditions = [f"{col}::text ILIKE %s" for col in search_columns]
+            conditions.append(f"({' OR '.join(search_conditions)})")
+            values += [f"%{search}%"] * len(search_columns)
+        
+        # Combine conditions for count query
+        count_query = f"SELECT COUNT(*) as total FROM {table_name}"
+        if conditions:
+            count_query += ' WHERE ' + ' AND '.join(conditions)
+        
+        # Execute count query
+        total_rows = get_data(table_name, filters, search, search_columns, columns=['COUNT(*) as total'], db_alias=db_alias)[0]["total"]
+        
+        # Calculate pagination metadata
+        total_pages = (total_rows + per_page - 1) // per_page
+        last_page = total_pages
+        
+        # Add conditions to main query
+        if conditions:
+            sql_query += ' WHERE ' + ' AND '.join(conditions)
+        
+        # Add ORDER BY clause
+        if order_by:
+            if isinstance(order_by, list):
+                order_by_clause = ', '.join(order_by)
+            else:
+                order_by_clause = order_by
+            sql_query += f" ORDER BY {order_by_clause}"
+        
+        # Add pagination
+        sql_query += f" LIMIT %s OFFSET %s"
+        values.extend([per_page, offset])
+        
+        # Execute paginated query
+        data = get_data(table_name, filters, search, search_columns, columns, limit=per_page, offset=offset, order_by=order_by, db_alias=db_alias)
+        
+        # Build pagination URLs
+        base_url = request.build_absolute_uri(request.path)
+        query_params = request.GET.copy()
+        
+        next_page_url = None
+        prev_page_url = None
+        
+        if page < total_pages:
+            query_params['current_page'] = str(page + 1)
+            next_page_url = f"{base_url}?{urlencode(query_params)}"
+        
+        if page > 1:
+            query_params['current_page'] = str(page - 1)
+            prev_page_url = f"{base_url}?{urlencode(query_params)}"
+        
+        # Construct response
+        pagination = {
+            "current_page": page,
+            "page_size": per_page,
+            "total_pages": total_pages,
+            "total_rows": total_rows,
+            "last_page": last_page,
+            "next_page_url": next_page_url,
+            "prev_page_url": prev_page_url,
+            "rows": data,
+        }
+        
+        return pagination
+    
+    except Exception as e:
+        raise Exception(f"Error in read with pagination from database '{db_alias}': {e}")
\ No newline at end of file
